/*! \file SphericalPowerDynamicMesh.h
*   \brief Spherical Power Diagram, Spherical Power Delaunay Mesh
*   \author David Gu
*   \date   documented on 11/17/2020
*
*   Dynamic Mesh for spherical power diagram and spherical power Delaunay triangulation
*/
#ifndef  _SPHERICAL_POWER_DYNAMIC_MESH_H_
#define  _SPHERICAL_POWER_DYNAMIC_MESH_H_

#include <map>
#include <vector>
#include <stack>
#include <Eigen/Eigen>

#include "Mesh/Vertex.h"
#include "Mesh/Edge.h"
#include "Mesh/Face.h"
#include "Mesh/HalfEdge.h"
#include "Mesh/DynamicMesh.h"
#include "Mesh/Boundary.h"
#include "Mesh/Iterators.h"
#include "Parser/parser.h"
#include "Parser/traits_io.h"
#include "Geometry/Polygon2D.h"
#include "Geometry/Polygon3D.h"
#include "Geometry/Plane3D.h"
#include "Geometry/Transformation.h"
#include "Geometry/predicates.h"

namespace MeshLib
{
	/*! \brief CPowerFace class
	*
	*	Face class for power diagram
	*   Trait: face normal
	*/
	class CPowerFace : public CFace
	{
	protected:
		/*! face normal */
		CPoint m_normal;
		/*! face area  */
		double m_area;
		/*! face dual point */
		CPoint m_dual_point;
		/*! inside the target domain */
		bool m_inside;
		/*! normal map*/
		CPoint m_normal_map;
		/*! 2D transformation */
		CTransformation2D m_2d_transformation;
		/*! 3D transformation */
		CTransformation   m_transformation;

	public:
		/*! face normal */
		CPoint& normal() { return m_normal; };
		/*! face area */
		double& area() { return m_area; };
		/*! face dual point*/
		CPoint& dual_point() { return m_dual_point; };
		/*! inside the target domain*/
		bool& inside() { return m_inside; };

		/*! 2D transformation */
		CTransformation2D& planar_transformation() { return m_2d_transformation; };
		/*! 3D transformation */
		CTransformation& transformation() { return m_transformation; };

	};

	/*! \brief CPowerVertex class
	*
	*   Vertex class for power diagram
	*   Trait : vertex rgb color
	*/
	class CPowerVertex : public  CVertex
	{
	protected:
		/*! vertex rgb color */
		CPoint   m_rgb;
		/*! vertex convex hull point, Kantorovich potential phi*/
		CPoint   m_phi;
		/*! dual cell in 3D*/
		CPolygon3D m_dual_cell3D;
		/*! normal map*/
		CPoint m_normal_map;
		/*! dual cell center */
		CPoint m_cell_center;

	public:
		/*! CViewerVertex Constructor */
		CPowerVertex()
		{
			m_rgb = CPoint(1, 1, 1); //default color is white
			m_normal = CPoint(1, 0, 0);
		}
		/*! vertex rgb color */
		CPoint& rgb() { return m_rgb; };
		/*! vertex Kantorovich potential phi*/
		CPoint& phi() { return m_phi; };

		/*! read vertex rgb, uv from vertex string */
		void _from_string();
		/*! write vertex uv to vertex string */
		void _to_string();

		/*! vertex conjudate cell */
		CPolygon3D& dual_cell3D() { return m_dual_cell3D; };

		/*! normal map*/
		CPoint& normal_map() { return m_normal_map; };
		
		/*! cell center */
		CPoint& cell_center() { return m_cell_center; };

	};

	// read vertex rgb, uv from vertex string 
	inline void CPowerVertex::_from_string()
	{
		CParser parser(m_string);

		for (std::list<CToken*>::iterator iter = parser.tokens().begin(); iter != parser.tokens().end(); ++iter)
		{
			CToken* token = *iter;
			if (token->m_key == "uv")
			{
				token->m_value >> m_uv;
			}
			else if (token->m_key == "rgb")
			{
				token->m_value >> m_rgb;
			}
			else if (token->m_key == "normal")
			{
				token->m_value >> m_normal_map;
			}
			else if (token->m_key == "u")
			{
				token->m_value >> m_phi;
			}

		}
	};

	// read vertex rgb, uv from vertex string 
	inline void CPowerVertex::_to_string()
	{
		CParser parser(m_string);

		parser._removeToken("uv");
		parser._toString(m_string);

		std::stringstream iss;
		iss << "uv=(" << m_uv[0] << " " << m_uv[1] << ") ";

		if (m_string.length() > 0)
		{
			m_string += " ";
		}
		m_string += iss.str();

	};

	/*! \brief CPowerEdge class
	*
	*   Edge class for power diagram
	*   Trait : Edge sharp
	*/
	class CPowerEdge : public  CEdge
	{
	protected:
		/*! edge sharp */
		bool   m_sharp;
		/*! processed */
		bool m_processed;
	public:
		/*! COTVertex Constructor */
		CPowerEdge()
		{
			m_sharp = false;
		}
		/*! vertex rgb color */
		bool& sharp() { return m_sharp; };
		/*! processed used in Lawson Edge Flip*/
		bool& processed() { return m_processed; };

		/*! read vertex rgb, uv from vertex string */
		void _from_string();
	};


	// read vertex rgb, uv from vertex string 
	inline void CPowerEdge::_from_string()
	{
		CParser parser(m_string);

		for (std::list<CToken*>::iterator iter = parser.tokens().begin(); iter != parser.tokens().end(); ++iter)
		{
			CToken* token = *iter;
			if (token->m_key == "sharp")
			{
				m_sharp = true;
			}
			break;
		}
	};

	/*! \brief CPowerHalfEdge class
	 *
	 *   HalfEdge class for power diagram
	 *   Trait : corner uv, corner id
	 */
	class CPowerHalfEdge : public  CHalfEdge
	{
	public:
		/*! corner uv */
		CPoint2& uv() { return m_uv; };
		/*! corner id*/
		int& cid() { return m_cid; };
		/*! spherical height */
		double& spherical_height() { return m_spherical_height; };
		/*! spherical dual height */
		double& spherical_dual_height() { return m_spherical_dual_height; };

	protected:
		/*! halfedge uv */
		CPoint2   m_uv;
		/*! halfedge corner id*/
		int m_cid;
		/*! spherical height*/
		double m_spherical_height;
		/*! spherical dual height */
		double m_spherical_dual_height;
	};

	/*-------------------------------------------------------------------------------------------------------------------------------------

		Power Diagram Mesh

	--------------------------------------------------------------------------------------------------------------------------------------*/
	/*! \brief CPowerDiagramMesh class
	*
	*	mesh class for optimal transport
	*
	*/
	template<typename V, typename E, typename F, typename H>
	class CSphericalPowerDiagramDynamicMesh : public CDynamicMesh<V, E, F, H>
	{
	public:

		typedef V CVertex;
		typedef E CEdge;
		typedef F CFace;
		typedef H CHalfEdge;

		typedef CBoundary<V, E, F, H> CBoundary;
		typedef CLoop<CVertex, CEdge, CFace, CHalfEdge> CLoop;
		typedef MeshVertexIterator<V, E, F, H> MeshVertexIterator;
		typedef MeshFaceIterator<V, E, F, H>   MeshFaceIterator;
		typedef MeshEdgeIterator<V, E, F, H>   MeshEdgeIterator;
		typedef VertexVertexIterator<V, E, F, H> VertexVertexIterator;
		typedef FaceVertexIterator<V, E, F, H> FaceVertexIterator;
		typedef VertexEdgeIterator<V, E, F, H> VertexEdgeIterator;
		typedef VertexFaceIterator<V, E, F, H> VertexFaceIterator;
		typedef VertexOutHalfedgeIterator<V, E, F, H> VertexOutHalfedgeIterator;
		typedef FaceHalfedgeIterator<V, E, F, H>  FaceHalfedgeIterator;

	protected:
		//compute the face dual point
		//each face is a plane z=ax+by-c, the dual point is (a,b,c)
			/*
			*	dual point of a face, Ax+By -C = Z, (A,B,C)
			*
			*   <p_k, (x,y,-1)> = z, k = 0, 1, 2
			*
			*	<p_1-p_0, (x,y,-1)> = 0,
			*   <p_2-p_0, (x,y,-1)> = 0,
			*
			*   (x,y,-1) parallel to the normal
			*/
		virtual void __face_dual_point(CFace* pf)
		{
			//prime function is phi

			CPoint p[3];
			int i = 0;
			for (FaceVertexIterator fviter(pf); !fviter.end(); fviter++)
			{
				CVertex* pv = *fviter;
				p[i++] = pv->phi();
			}
			CPoint n = (p[1] - p[0]) ^ (p[2] - p[0]);
			n /= n.norm();
			pf->normal() = n;

			double t[3];
			for (int j = 0; j < 3; j++)
			{
				t[j] = 1.0 / (n * p[j]);
			}

			pf->dual_point() = n * ((t[0] + t[1] + t[2]) / 3.0);
		};

		//compute power cell 
		virtual void __power_cell_area(CVertex* pv)
		{
			std::vector<CPoint>  points;

			for (VertexFaceIterator vfiter(pv); !vfiter.end(); vfiter++)
			{
				CFace* pf = *vfiter;
				points.push_back(pf->dual_point());
			}

			CPolygon3D Poly;
			for (int i = 0; i < (int)points.size(); i++)
			{
				CSegment3D seg(points[i], points[(i + 1) % points.size()]);
				Poly.add(seg);
			}
			pv->dual_cell3D() = Poly;
			//need to compute
			pv->dual_area() = Poly.spherical_project().area();
			//
			if (std::isnan(pv->dual_area()))
			{
				printf("[WARNING] dual area nan!\n");
				pv->dual_area() = 0;
			}
		};	

		//verify if an edge is local power Delaunay
		virtual bool __edge_local_convex(CEdge* pe)
		{
			double p[4][3];
			CVertex* pV[4];

			CHalfEdge* ph = this->edgeHalfedge(pe, 0);
			pV[0] = this->halfedgeSource(ph);
			pV[1] = this->halfedgeTarget(ph);
			ph =	this->halfedgeNext(ph);
			pV[2] = this->halfedgeTarget(ph);
			ph = this->edgeHalfedge(pe, 1);
			ph = this->halfedgeNext(ph);
			pV[3] = this->halfedgeTarget(ph);

			for (int k = 0; k < 4; k++)
			{
				CPoint& pt = pV[k]->phi();
				for (int i = 0; i < 3; i++)
				{
					p[k][i] = pt[i];
				}
			}

			// worst OT
			double sign = orient3d(p[0], p[1], p[2], p[3]);
			if (sign < 0)
				return false;
			return true;
		}
		//verify if an edge is flippable
		virtual bool __edge_flippable(CEdge* pe)
		{
			CVertex* pV[4];

			CHalfEdge* ph = this->edgeHalfedge(pe, 0);
			pV[0] = this->halfedgeSource(ph);
			pV[1] = this->halfedgeTarget(ph);
			ph =	this->halfedgeNext(ph);
			pV[2] = this->halfedgeTarget(ph);
			ph =	this->edgeHalfedge(pe, 1);
			ph =	this->halfedgeNext(ph);
			pV[3] = this->halfedgeTarget(ph);

			for (VertexVertexIterator vviter(pV[2]); !vviter.end(); vviter++)
			{
				CVertex* pW = *vviter;
				if (pW == pV[3]) return false;
			}
			return true;
		};


	public:

		/*!----------------------------------------------------------------------------------------------
		 * methods to compute the convex hull, Power Delaunay triangulation
		-----------------------------------------------------------------------------------------------*/

		//edge swap to compute convex hull
		bool _edge_swap()
		{
			int flipped_edges = 0;
			int non_flipplable = 0;

			while (true)
			{
				flipped_edges = 0;
				non_flipplable = 0;

				for (MeshEdgeIterator eiter(this); !eiter.end(); eiter++)
				{
					CEdge* pe = *eiter;
					//if (this->isBoundary(pe))
					//skip all the edges connecting to the infinity vertex
					//skill the link of the infinity vertex
					if (_boundary(pe) || _infinite(pe)) continue;

					if (!__edge_local_convex(pe))
					{
						if (__edge_flippable(pe))
						{
							swapEdge(pe);
							flipped_edges++;
						}
						else
							non_flipplable++;
					}
				}

				if (flipped_edges == 0) break;
			}

			if (non_flipplable > 0)
			{
				std::cout << "Nonconvex" << std::endl;
				return false;
			}
			std::cout << "Convex" << std::endl;
			return true;
		}
		//Charles Lawson algorithm, edge swap for nearest and farthest Delaunay Triangulation
		bool _Lawson_edge_swap()
		{
			int flipped_edges = 0;
			int non_flipplable = 0;

			//std::queue<CEdge*> illegal_edges;
			std::stack<CEdge*> illegal_edges;

			//unmark all edges
			for (MeshEdgeIterator eiter(this); !eiter.end(); eiter++)
			{
				CEdge* pe = *eiter;
				pe->processed() = false;
			}

			//enqueue illegal edges
			for (MeshEdgeIterator eiter(this); !eiter.end(); eiter++)
			{
				CEdge* pe = *eiter;
				if (!__edge_local_convex(pe))
				{
					pe->processed() = true;
					illegal_edges.push(pe);
				}
			}
			//process all illegal edges
			while (!illegal_edges.empty())
			{
				//CEdge* pe = illegal_edges.front();
				CEdge* pe = illegal_edges.top();
				illegal_edges.pop();
				pe->processed() = false;

				if (!__edge_local_convex(pe))
				{
					if (__edge_flippable(pe))
					{
						this->swapEdge(pe);

						std::vector<CEdge*> nbh_edges;

						CHalfEdge* ph = this->edgeHalfedge(pe, 0);
						nbh_edges.push_back(this->halfedgeEdge(this->halfedgeNext(ph)));
						nbh_edges.push_back(this->halfedgeEdge(this->halfedgePrev(ph)));
						CHalfEdge* ps = this->edgeHalfedge(pe, 1);
						nbh_edges.push_back(this->halfedgeEdge(this->halfedgeNext(ps)));
						nbh_edges.push_back(this->halfedgeEdge(this->halfedgePrev(ps)));

						for (size_t k = 0; k < nbh_edges.size(); k++)
						{
							CEdge* pE = nbh_edges[k];
							if (pE->processed()) continue;
							pE->processed() = true;
							illegal_edges.push(pE);
						}
					}
				}
			}

			//enqueue illegal edges
			for (MeshEdgeIterator eiter(this); !eiter.end(); eiter++)
			{
				CEdge* pe = *eiter;
				if (!__edge_local_convex(pe))
				{
					std::cout << "Non Convex" << std::endl;
					return false;
				}
			}

			return true;
		}

		/*!----------------------------------------------------------------------------------------------
		 * methods to compute upper envelop, Power Voronoi Diagram
		-----------------------------------------------------------------------------------------------*/

		//Legendre transform
		virtual void _Legendre_transform()
		{

			for (MeshFaceIterator fiter(this); !fiter.end(); fiter++)
			{
				CFace* pf = *fiter;
				__face_dual_point(pf);
			}
		};

		/*!----------------------------------------------------------------------------------------------
		 * methods to clip power cells
		-----------------------------------------------------------------------------------------------*/

		//clip the power cells by D
		virtual void _power_cell_area()
		{

			double total_area = 0;
			for (MeshVertexIterator viter(this); !viter.end(); viter++)
			{
				CVertex* pv = *viter;
				__power_cell_area(pv);
				total_area += pv->dual_area();
			}
			std::cout << "Total dual area is " << total_area << std::endl;
		}
		//compute the mass center of the power cells
		virtual void _power_cell_center()
		{
			for (MeshVertexIterator viter(this); !viter.end(); viter++)
			{
				CVertex* pv = *viter;
				if (pv->dual_cell3D().edges().empty()) {
					std::cout << "Warning: dual cell empty" << pv->id() << std::endl;
					continue;
				}
				pv->cell_center() = pv->dual_cell3D().mass_center();
			}
		}


		/*!
		*	compute halfedge spherical distance from the power circle center to the current edge
		*/
		void __halfedge_spherical_height()
		{
			for (MeshFaceIterator fiter(this); !fiter.end(); fiter++)
			{
				CFace* pf = *fiter;
				for (FaceHalfedgeIterator fhiter(pf); !fhiter.end(); fhiter++)
				{
					CHalfEdge* ph = *fhiter;
					CHalfEdge* ps =  this->halfedgeSym(ph);
					CFace* pLf = this->halfedgeFace(ph);
					CFace* pRf = this->halfedgeFace(ps);
					CVertex* pT = this->halfedgeTarget(ph);
					CVertex* pS = this->halfedgeSource(ph);

					CPoint T = pT->phi();
					CPoint S = pS->phi();
					CPoint OL = pLf->normal();
					CPoint OR = pRf->normal();

					T /= T.norm();
					S /= S.norm();

					CPoint n1 = S ^ T;
					CPoint n2 = OL ^ OR;
					n2 /= n2.norm();
					n1 /= n1.norm();
					CPoint foot = n2 ^ n1;
					foot /= foot.norm();

					double s = ((S - OL) ^ (T - OL)) * OL;
					double d = acos(foot * OL);
					ph->spherical_height() = (s > 0) ? d : -d;

					//compute dual height
					{
						CPoint n = S ^ T;
						n /= n.norm();

						CPoint d = foot^ T;
						ph->spherical_dual_height() = asin(d * n);
					}
				}
			}
		}
		/*!
		*	compute face spherical power circle center, which equals to the face normal
		*/
		void __face_power_center()
		{
			for (MeshFaceIterator fiter(this); !fiter.end(); fiter++)
			{
				CFace* pf = *fiter;
				std::vector<CPoint> pts;
				for (FaceVertexIterator fviter(pf); !fviter.end(); fviter++)
				{
					CVertex* pv = *fviter;
					pts.push_back(pv->phi());
				}
				CPoint n = (pts[1] - pts[0]) ^ (pts[2] - pts[0]);
				n /= n.norm();
				pf->normal() = n;
			}
		}

		void _edge_interior_weight(CEdge* pe)
		{

			CHalfEdge* ph1 = this->edgeHalfedge(pe, 0);
			CHalfEdge* ph2 = this->edgeHalfedge(pe, 1);
			{
				double dl = ph1->spherical_height();
				double dr = ph2->spherical_height();
				double alpha = ph1->spherical_dual_height();
				double beta = ph2->spherical_dual_height();
				double w = -(sin(dl) + sin(dr)) / (tan(alpha) + tan(beta));
				//std::cout << fabs(pe->weight() - w) << " ";
				pe->weight() = w;
			}
		};
		/*!
		*	compute edge weight
		*/
		void _edge_weight()
		{
			//the order matters
			__face_power_center();
			__halfedge_spherical_height();
			for (MeshEdgeIterator eiter(this); !eiter.end(); eiter++)
			{
				CEdge* pe = *eiter;
				_edge_interior_weight(pe);
			}
		};


	};

	typedef CSphericalPowerDiagramDynamicMesh<CPowerVertex, CPowerEdge, CPowerFace, CPowerHalfEdge> CPDMesh;


}
#endif  //! _SPHERICAL_POWER_DYNAMIC_MESH_H_

